"""Library for manipulating on-screen text using gaze tracking and OCR."""

import os.path
import time
from concurrent import futures

import dragonfly

from . import _dragonfly_wrappers as dragonfly_wrappers


class Controller(object):
    """Mediates interaction with gaze tracking and OCR."""

    def __init__(self,
                 ocr_reader,
                 eye_tracker,
                 save_data_directory=None,
                 mouse=dragonfly_wrappers.Mouse(),
                 keyboard=dragonfly_wrappers.Keyboard()):
        self.ocr_reader = ocr_reader
        self.eye_tracker = eye_tracker
        self.save_data_directory = save_data_directory
        self.mouse = mouse
        self.keyboard = keyboard
        self._executor = futures.ThreadPoolExecutor(max_workers=1)
        self._future = None

    def shutdown(self, wait=True):
        self._executor.shutdown(wait)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown(wait=True)
        return False

    def start_reading_nearby(self):
        """Start OCR nearby the gaze point in a background thread."""
        gaze_point = self.eye_tracker.get_gaze_point_or_default()
        # Don't enqueue multiple requests.
        if self._future and not self._future.done():
            self._future.cancel()
        self._future = self._executor.submit(lambda: self.ocr_reader.read_nearby(gaze_point))

    def latest_screen_contents(self):
        """Return the ScreenContents of the latest call to start_reading_nearby().

        Blocks until available.
        """
        if not self._future:
            raise RuntimeError("Call start_reading_nearby() before latest_screen_contents()")
        return self._future.result()

    def move_cursor_to_word(self, word, cursor_position="middle"):
        """Move the mouse cursor nearby the specified word.

        If successful, returns the new cursor coordinates.

        Arguments:
        word: The word to search for.
        cursor_position: "before", "middle", or "after" (relative to the matching word)
        """
        screen_contents = self.latest_screen_contents()
        coordinates = screen_contents.find_nearest_word_coordinates(word, cursor_position)
        self._write_data(screen_contents, word, coordinates)
        if coordinates:
            self.mouse.move(coordinates)
            return coordinates
        else:
            return False

    def move_text_cursor_to_word(self, word, cursor_position="middle", use_nearest=True,
                                 validate_location_function=None):
        """Move the mouse cursor nearby the specified word.

        If successful, returns the new cursor coordinates.

        Arguments:
        word: The word to search for.
        cursor_position: "before", "middle", or "after" (relative to the matching word).
        use_nearest: Minimizes cursor movement for subword placement, instead of always
                     clicking based on cursor_position.
        validate_location_function: Given a word location, return whether to proceed with
                                    cursor movement.
        """
        screen_contents = self.latest_screen_contents()
        word_location = screen_contents.find_nearest_word(word)
        self._write_data(screen_contents, word, word_location)
        if (not word_location or
            (validate_location_function and not validate_location_function(word_location))):
            return False
        if cursor_position == "before":
            if (not use_nearest
                or word_location.left_char_offset
                <= word_location.right_char_offset + len(word_location.text)):
                self.mouse.move(word_location.start_coordinates)
                self.mouse.click()
                if word_location.left_char_offset:
                    self.keyboard.right(word_location.left_char_offset)
            else:
                self.mouse.move(word_location.end_coordinates)
                self.mouse.click()
                offset = word_location.right_char_offset + len(word_location.text)
                if offset:
                    self.keyboard.left(offset)
        elif cursor_position == "middle":
            # Note: if it's helpful, we could change this to position the cursor
            # in the middle of the word.
            self.mouse.move(word_location.middle_coordinates)
            self.mouse.click()
        if cursor_position == "after":
            if (not use_nearest
                or word_location.right_char_offset
                <= word_location.left_char_offset + len(word_location.text)):
                self.mouse.move(word_location.end_coordinates)
                self.mouse.click()
                if word_location.right_char_offset:
                    self.keyboard.left(word_location.right_char_offset)
            else:
                self.mouse.move(word_location.start_coordinates)
                self.mouse.click()
                offset = word_location.left_char_offset + len(word_location.text)
                if offset:
                    self.keyboard.right(offset)
        return word_location

    def select_text(self, start_word, end_word=None):
        """Select a range of onscreen text.

        If only start_word is provided, it can be a word or phrase to select. If
        end_word is provided, a range from the start word to end word will be
        selected.
        """
        # Automatically split up start word if multiple words are provided.
        if " " in start_word and not end_word:
            words = start_word.split()
            start_word = words[0]
            end_word = words[-1]
        # Always click before the word to avoid subword selection issues on Windows.
        start_location = self.move_text_cursor_to_word(start_word, "before", use_nearest=False)
        if not start_location:
            return False
        if end_word:
            # If gaze has significantly moved, look for the end word at the final gaze coordinates.
            current_gaze_point = self.eye_tracker.get_gaze_point_or_default()
            previous_gaze_point = self.latest_screen_contents().screen_coordinates
            if (_distance_squared(current_gaze_point, previous_gaze_point)
                > _squared(self.ocr_reader.radius / 2.0)):
                self.start_reading_nearby()
        else:
            end_word = start_word
        self.keyboard.shift_down()
        validate_function = lambda location: self._is_valid_selection(start_location.start_coordinates,
                                                                      location.end_coordinates)
        # Always click after the word to avoid subword selection issues on Windows.
        end_location = self.move_text_cursor_to_word(
            end_word, "after", use_nearest=False,
            validate_location_function=validate_function)
        self.keyboard.shift_up()
        return end_location

    def move_cursor_to_word_action(self, word, cursor_position="middle"):
        """Return a Dragonfly action for moving the mouse cursor nearby a word."""
        outer = self
        class MoveCursorToWordAction(dragonfly.ActionBase):
            def _execute(self, data=None):
                dynamic_word = word
                if data:
                    dynamic_word = word % data
                return outer.move_cursor_to_word(dynamic_word, cursor_position)
        return MoveCursorToWordAction()

    def move_text_cursor_action(self, word, cursor_position="middle"):
        """Return a dragonfly action for moving the text cursor nearby a word."""
        outer = self
        class MoveTextCursorAction(dragonfly.ActionBase):
            def _execute(self, data=None):
                dynamic_word = word
                if data:
                    dynamic_word = word % data
                return outer.move_text_cursor_to_word(dynamic_word, cursor_position)
        return MoveTextCursorAction()

    def select_text_action(self, start_word, end_word=None):
        """Return a Dragonfly action for selecting text."""
        outer = self
        class SelectTextAction(dragonfly.ActionBase):
            def _execute(self, data=None):
                dynamic_start_word = start_word
                dynamic_end_word = end_word
                if data:
                    dynamic_start_word = start_word % data
                    if end_word:
                        try:
                            dynamic_end_word = end_word % data
                        except KeyError:
                            dynamic_end_word = None
                return outer.select_text(dynamic_start_word, dynamic_end_word)
        return SelectTextAction()

    def _write_data(self, screen_contents, word, word_location):
        if not self.save_data_directory:
            return
        file_name_prefix = "{}_{:.2f}".format("success" if word_location else "failure", time.time())
        file_path_prefix = os.path.join(self.save_data_directory, file_name_prefix)
        screen_contents.screenshot.save(file_path_prefix + ".png")
        with open(file_path_prefix + ".txt", "w") as file:
            file.write(word)

    def _is_valid_selection(self, start_coordinates, end_coordinates):
        epsilon = 5  # pixels
        (start_x, start_y) = start_coordinates
        (end_x, end_y) = end_coordinates
        # Selection goes to previous line.
        if end_y - start_y < -epsilon:
            return False
        # Selection stays on same line.
        elif end_y - start_y < epsilon:
            return end_x > start_x
        # Selection moves to following line.
        else:
            return True


def _squared(x):
    return x * x


def _distance_squared(coordinate1, coordinate2):
    return (_squared(coordinate1[0] - coordinate2[0])
            + _squared(coordinate1[1] - coordinate2[1]))



import dragonfly


class Mouse(object):
    def move(self, coordinates):
        dragonfly.Mouse("[{}, {}]".format(*coordinates)).execute()

    def click(self):
        dragonfly.Mouse("left").execute()

    def click_down(self):
        dragonfly.Mouse("left:down").execute()

    def click_up(self):
        dragonfly.Mouse("left:up").execute()

    def scroll_down(self, n=1):
        dragonfly.Mouse("wheeldown:{}".format(n)).execute()

    def scroll_up(self, n=1):
        dragonfly.Mouse("wheelup:{}".format(n)).execute()


class Keyboard(object):
    def type(self, text):
        dragonfly.Text(text.replace("%", "%%")).execute()

    def shift_down(self):
        dragonfly.Key("shift:down").execute()

    def shift_up(self):
        dragonfly.Key("shift:up").execute()

    def left(self, n=1):
        dragonfly.Key("left:{}".format(n)).execute()

    def right(self, n=1):
        dragonfly.Key("right:{}".format(n)).execute()


class Windows(object):
    def get_monitor_size(self):
        primary = dragonfly.Monitor.get_all_monitors()[0]
        return (primary.rectangle.dx, primary.rectangle.dy)
    
    def get_foreground_window_center(self):
        window_position = dragonfly.Window.get_foreground().get_position()
        return (window_position.x_center, window_position.y_center)



"""Tobii eye tracker wrapper."""

import math
import sys

from . import _dragonfly_wrappers as dragonfly_wrappers


class EyeTracker(object):
    _instance = None

    @classmethod
    def get_connected_instance(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = cls(*args, **kwargs)
        if not cls._instance.is_connected:
            cls._instance.connect()
        return cls._instance

    def __init__(self,
                 tobii_dll_directory,
                 mouse=dragonfly_wrappers.Mouse(),
                 keyboard=dragonfly_wrappers.Keyboard(),
                 windows=dragonfly_wrappers.Windows()):
        self._mouse = mouse
        self._keyboard = keyboard
        self._windows = windows
        # Attempt to load eye tracker DLLs.
        global clr, Action, Double, Host, GazeTracking
        try:
            import clr
            from System import Action, Double
            sys.path.append(tobii_dll_directory)
            clr.AddReference("Tobii.Interaction.Model")
            clr.AddReference("Tobii.Interaction.Net")
            from Tobii.Interaction import Host
            from Tobii.Interaction.Framework import GazeTracking
            self.is_mock = False
        except:
            print("Eye tracking libraries are unavailable.")
            self.is_mock = True
        self._host = None
        self._gaze_point = None
        self._gaze_state = None
        self._screen_scale = (1.0, 1.0)
        self._head_rotation = None
        self.is_connected = False

    def connect(self):
        if self.is_mock:
            return
        self._host = Host()

        # Connect handlers.
        screen_bounds_state = self._host.States.CreateScreenBoundsObserver()
        screen_bounds_state.Changed += self._handle_screen_bounds
        gaze_state = self._host.States.CreateGazeTrackingObserver()
        gaze_state.Changed += self._handle_gaze_state
        gaze_points = self._host.Streams.CreateGazePointDataStream()
        action = Action[Double, Double, Double](self._handle_gaze_point)
        gaze_points.GazePoint(action)
        head_pose = self._host.Streams.CreateHeadPoseStream()
        head_pose.Next += self._handle_head_pose
        self.is_connected = True
        print("Eye tracker connected.")

    def disconnect(self):
        if not self.is_connected:
            return
        self._host.DisableConnection()
        self._host = None
        self._gaze_point = None
        self._gaze_state = None
        self.is_connected = False
        print("Eye tracker disconnected.")

    def _handle_screen_bounds(self, sender, state):
        if not state.IsValid:
            print("Ignoring invalid screen bounds.")
            return
        bounds = state.Value
        monitor_size = self._windows.get_monitor_size()
        self._screen_scale = (monitor_size[0] / float(bounds.Width),
                              monitor_size[1] / float(bounds.Height))

    def _handle_gaze_state(self, sender, state):
        if not state.IsValid:
            print("Ignoring invalid gaze state.")
            return
        self._gaze_state = state.Value

    def _handle_gaze_point(self, x, y, timestamp):
        self._gaze_point = (x, y, timestamp)

    def _handle_head_pose(self, sender, stream_data):
        pose = stream_data.Data
        self._head_rotation = (pose.HeadRotation.X,
                               pose.HeadRotation.Y,
                               pose.HeadRotation.Z)

    def has_gaze_point(self):
        return (not self.is_mock and
                self._gaze_state == GazeTracking.GazeTracked and
                self._gaze_point)

    def get_gaze_point_or_default(self):
        if self.has_gaze_point():
            return (self._gaze_point[0] * self._screen_scale[0],
                    self._gaze_point[1] * self._screen_scale[1])
        else:
            return self._windows.get_foreground_window_center()

    def print_gaze_point(self):
        if not self.has_gaze_point():
            print("No valid gaze point.")
            return
        print("Gaze point: (%f, %f)" % self._gaze_point[:2])

    def move_to_gaze_point(self, offset=(0, 0)):
        gaze = self.get_gaze_point_or_default()
        x = max(0, int(gaze[0]) + offset[0])
        y = max(0, int(gaze[1]) + offset[1])
        self._mouse.move((x, y))

    def type_gaze_point(self, format):
        self._keyboard.type(format % self.get_gaze_point_or_default()).execute()

    def get_head_rotation_or_default(self):
        rotation = self._head_rotation or (0, 0, 0)
        if math.isnan(rotation[0]):
            rotation = (0, 0, 0)
        return rotation



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gaze-ocr",
    version="0.2.0",
    author="James Stout",
    author_email="james.wolf.stout@gmail.com",
    description="Library for applying OCR to where the user is looking.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wolfmanstout/gaze-ocr",
    packages=["gaze_ocr"],
    install_requires=[
        "screen-ocr",
        "dragonfly2",
        "pythonnet",
        'futures; python_version < "3.2"',
    ],
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)
